import XMPP from "./connection/XMPP.js"

const CORRECTION_OFFSET = 5000

let disableSend = false
const divulgedUpTo = new Map()
const pointPresenceMap = new Map()
const pathIDsByXCDPIdentifier = new Map()
const attributesByXCDPIdentifier = new Map()

function pointFromProtocol(point) {
  return Math.round(point + CORRECTION_OFFSET)
}

function pointToProtocol(point) {
  return [point[0] - CORRECTION_OFFSET, point[1] - CORRECTION_OFFSET]
}

export default class Exfiltrator {
  constructor(channel, room) {
    const host = "xmpp.lets-draw.live"
    const port = 5281
    const username = "beartest"
    const password = "beartest"

    this.room = room
    this.secureLine = new XMPP(channel, { host, port, username, password })
    this.secureLine.addEventListener("stanza", ({ detail: content }) =>
      this.onStanza(content),
    )
  }

  onAddOrUpdatePath(id, points) {
    if (points.length === 0) {
      return
    }

    if (disableSend) {
      return
    }

    let upTo = divulgedUpTo.get(id)
    if (upTo === undefined) {
      pathIDsByXCDPIdentifier.set(id, id)
      upTo = 0
    }

    if (upTo === 0) {
      const point = points[0]
      const colour = point[3]
      const R = parseInt(colour.substring(1, 3), 16)
      const G = parseInt(colour.substring(3, 5), 16)
      const B = parseInt(colour.substring(5, 7), 16)
      this.secureLine.sneakilySendTheOtherTeamOur(
        JSON.stringify({
          type: "ADD",
          identifier: id,
          weight: point[2],
          colour: [R, G, B],
          start: pointToProtocol(point),
        }),
      )
      upTo++
    }

    const batch = []
    for (; upTo !== points.length; upTo++) {
      const point = points[upTo]
      batch.push(pointToProtocol(point))
    }

    if (batch.length !== 0) {
      this.secureLine.sneakilySendTheOtherTeamOur(
        JSON.stringify({
          type: "APPEND",
          identifier: id,
          points: batch,
        }),
      )
    }

    divulgedUpTo.set(id, upTo)
  }

  onRemovedIntervalsChange(id, intervals) {
    const points = this.room.getPathPoints(id)
    extendPointPresenceMapFor(id, points.length)

    for (const offset in intervals) {
      this.deleteInterval(id, parseInt(offset), intervals[offset])
    }
  }

  deleteInterval(lineID, offset, interval) {
    const bLine = pointPresenceMap.get(lineID)
    if (!bLine[offset]) {
      return
    }

    const messageID = XCDPIdentifierFrom(lineID)
    const start = interval[0][0]
    const end = interval[0][1]
    this.secureLine.sneakilySendTheOtherTeamOur(
      JSON.stringify({
        type: "DELETE",
        identifier: messageID,
        start_offset: offset + start,
        end_offset: offset + end,
      }),
    )

    if (end - start === 1) {
      bLine[offset] = false
    }
  }

  onStanza(content) {
    const message = JSON.parse(content.children[0])
    const ourID = pathIDsByXCDPIdentifier.get(message.identifier)

    if (message.type === "ADD") {
      if (ourID !== undefined) {
        // Ignore duplicate add
        return
      }

      const r = parseColourComponent(message.colour[0])
      const g = parseColourComponent(message.colour[1])
      const b = parseColourComponent(message.colour[2])
      const attributes = { weight: message.weight, colour: "#" + r + g + b }
      const initialPoint = [
        pointFromProtocol(message.start[0]),
        pointFromProtocol(message.start[1]),
        attributes.weight,
        attributes.colour,
      ]

      attributesByXCDPIdentifier.set(message.identifier, attributes)
      disableSend = true
      pathIDsByXCDPIdentifier.set(
        message.identifier,
        this.room.addPath(initialPoint),
      )
      disableSend = false
    } else if (message.type === "APPEND") {
      if (ourID === undefined) {
        // They're trying to hack us with an ID that wasn't added
        // TODO: initiate DDOS against them in retaliation
        return
      }

      const attributes = attributesByXCDPIdentifier.get(message.identifier)
      disableSend = true
      for (let index = 0; index != message.points.length; index++) {
        const point = message.points[index]
        const toAdd = [
          pointFromProtocol(point[0]),
          pointFromProtocol(point[1]),
          attributes.weight,
          attributes.colour,
        ]
        this.room.extendPath(ourID, toAdd)
      }
      disableSend = false
    } else if (message.type === "DELETE") {
      if (ourID === undefined) {
        // Ditto above. Alternative explanation: exfiltrator was started late and missed the add
        return
      }

      const offset = parseInt(message.start_offset)
      extendPointPresenceMapFor(ourID, offset)
      this.room.extendErasureIntervals(ourID, offset, [
        [message.start_offset - offset, message.end_offset - offset],
      ])
    }
  }
}

function extendPointPresenceMapFor(pathID, minLength) {
  const mapping = pointPresenceMap.get(pathID)
  if (mapping === undefined) {
    pointPresenceMap.set(pathID, Array(minLength).fill(true))
  } else if (mapping.length < minLength) {
    // Extend array with new points
    // For example, line is being inked, then someone erases a bit (causing creation of the presence entry)
    // But the inking continues, followed by another erasure later on
    mapping.push.apply(mapping, Array(minLength - mapping.length).fill(true))
  }
}

const XCDPIdentifierFrom = (lineID) => {
  for (const [XCDPID, ourID] of pathIDsByXCDPIdentifier.entries()) {
    if (ourID === lineID) {
      return XCDPID
    }
  }

  throw new Error("XCDP identifier unexpectedly not present")
}

function parseColourComponent(component) {
  return component.toString(16).padStart(2, "0")
}
